﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows;
using System.Diagnostics;
using System.Windows.Media;

namespace SeeMoreStackPanelTestApplication
{
    public class SeeMoreStackPanel : VirtualizingPanel, IScrollInfo
    {
        #region IScrollInfo
        public bool CanHorizontallyScroll { get; set; }

        public bool CanVerticallyScroll { get { return true; } set { } }

        private Size extentSize = new Size();
        public double ExtentHeight
        {
            get { return extentSize.Height; }
        }

        public double ExtentWidth
        {
            get { return extentSize.Width; }
        }

        private Size viewportSize = new Size();
        public double ViewportHeight
        {
            get { return viewportSize.Height; }
        }

        public double ViewportWidth
        {
            get { return viewportSize.Width; }
        }

        public double HorizontalOffset
        {
            get;
            private set;
        }

        public double VerticalOffset
        {
            get;
            private set;
        }

        public void SetHorizontalOffset(double offset)
        {

        }

        public void SetVerticalOffset(double offset)
        {
            if (offset < 0 || viewportSize.Height >= ExtentHeight)
            {
                offset = 0;
            }
            else
            {
                if (offset + viewportSize.Height >= ExtentHeight)
                {
                    offset = ExtentHeight - viewportSize.Height;
                }
            }
            VerticalOffset = offset;

            ScrollOwner.InvalidateScrollInfo();
            InvalidateMeasure();
        }

        public void LineDown()
        {
            SetVerticalOffset(VerticalOffset + 10);
        }

        public void LineUp()
        {
            SetVerticalOffset(VerticalOffset - 10);
        }

        public void LineLeft()
        {
            throw new NotImplementedException();
        }

        public void LineRight()
        {
            throw new NotImplementedException();
        }

        public System.Windows.Rect MakeVisible(System.Windows.Media.Visual visual, System.Windows.Rect rectangle)
        {
            return new Rect();
        }

        public void MouseWheelDown()
        {
            SetVerticalOffset(VerticalOffset + 10);
        }

        public void MouseWheelUp()
        {
            SetVerticalOffset(VerticalOffset - 10);
        }

        public void MouseWheelLeft()
        {
            throw new NotImplementedException();
        }

        public void MouseWheelRight()
        {
            throw new NotImplementedException();
        }

        public void PageDown()
        {
            SetVerticalOffset(VerticalOffset + 30);
        }

        public void PageUp()
        {
            SetVerticalOffset(VerticalOffset - 30);
        }

        public void PageLeft()
        {
            throw new NotImplementedException();
        }

        public void PageRight()
        {
            throw new NotImplementedException();
        }

        public ScrollViewer ScrollOwner
        {
            get;
            set;
        }
        #endregion

        protected override void OnItemsChanged(object sender, ItemsChangedEventArgs args)
        {
            if (args.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Reset)
            {
                InvalidateMeasure();
            }
        }



        protected override Size MeasureOverride(Size availableSize)
        {
            ItemsControl itemsControl = ItemsControl.GetItemsOwner(this);
            if (itemsControl == null)
            {
                throw new InvalidOperationException("단독으로 사용될 수 없는 Panel입니다.");
            }
            UIElementCollection children = this.InternalChildren;

            //새 ItemContainer를 생성할 경우는 다음 두가지 경우 이다. 
            //첫번째는 처음 로딩되었을 경우이며
            //두번째는 ScrollToEnd가 동작되었을 경우이다.
            if ((VerticalOffset == 0 && InternalChildren.Count == 0) || VerticalOffset + ViewportHeight == ExtentHeight)
            {
                CreateItemContainer(availableSize);
            }

            this.viewportSize = availableSize;
            this.ScrollOwner.InvalidateScrollInfo();
            return availableSize;
        }

        private void CreateItemContainer(Size availableSize)
        {
            //처음 ItemsControl이 로드되었을 경우
            if (VerticalOffset == 0 && InternalChildren.Count == 0)
            {
                double remainHeight = availableSize.Height;
                GeneratorPosition position = new GeneratorPosition(-1, 0);
                IItemContainerGenerator generator = ItemContainerGenerator as IItemContainerGenerator;
                using (generator.StartAt(position, GeneratorDirection.Forward, true))
                {
                    while (remainHeight > 0)
                    {
                        bool isRealized;
                        UIElement element = generator.GenerateNext(out isRealized) as UIElement;
                        if (element == null) break;
                        if (isRealized) //새로 생성된 ItemContainer
                        {
                            Debug.Assert(InternalChildren.Contains(element) == false, "새로 얻은 element이므로 InternalChildren은 이 element를 포함하지 않는다.");
                            this.AddInternalChild(element);
                            ItemContainerGenerator.PrepareItemContainer(element);
                            element.Measure(availableSize);
                        }
                        else
                        {
                            Debug.Assert(InternalChildren.Contains(element));
                        }
                        remainHeight -= element.DesiredSize.Height;
                    }
                }
                extentSize = availableSize;
                if (remainHeight < 0)
                {
                    this.extentSize.Height += remainHeight * -1;
                }
            }
            else //ScrollToEnd 명령이 불렸을 경우이다. 이 때는 5개의 Item을 추가적으로 생성하기로 한다.
            {
                Debug.Assert(VerticalOffset + viewportSize.Height == ExtentHeight);

                //모든 Item들이 추가되었을 경우
                if (InternalChildren.Count == ItemsControl.GetItemsOwner(this).Items.Count)
                    return;

                IItemContainerGenerator generator = ItemContainerGenerator as IItemContainerGenerator;
                GeneratorPosition position = generator.GeneratorPositionFromIndex(InternalChildren.Count);
#if DEBUG
                int count = InternalChildren.Count;
#endif
                using (generator.StartAt(position, GeneratorDirection.Forward, true))
                {
                    for (int i = 0; i < 5; i++)
                    {
                        bool isRealized;
                        UIElement element = generator.GenerateNext(out isRealized) as UIElement;
                        //더 이상 생성할 Item이 없을 경우
                        if (element == null) break;
                        if (isRealized) //새로 생성된 ItemContainer
                        {
                            Debug.Assert(InternalChildren.Contains(element) == false, "새로 얻은 element이므로 InternalChildren은 이 element를 포함하지 않는다.");
                            this.AddInternalChild(element);
                            ItemContainerGenerator.PrepareItemContainer(element);
                            element.Measure(availableSize);
                            //ScrollViewer의 ExtentSize를 조절한다.
                            this.extentSize.Height += element.DesiredSize.Height;
                        }
                    }
                }

#if DEBUG
                Debug.Assert(ItemsControl.GetItemsOwner(this).Items.Count == InternalChildren.Count ||
                            count + 5 == InternalChildren.Count);
#endif
            }
        }

        TranslateTransform translateTransform = new TranslateTransform();
        public SeeMoreStackPanel()
        {
            this.RenderTransform = translateTransform;
        }

        protected override Size ArrangeOverride(Size finalSize)
        {
            ItemsControl itemsControl = ItemsControl.GetItemsOwner(this);
            if (itemsControl == null)
            {
                throw new InvalidOperationException("단독으로 사용될 수 없는 Panel입니다.");
            }
            double topPosition = 0;
            //InnerContainer에 추가된 Item들을 배치한다. 
            foreach (UIElement element in this.InternalChildren)
            {
                element.Arrange(new Rect(new Point(0, topPosition), element.DesiredSize));
                topPosition += element.RenderSize.Height;
            }
            translateTransform.Y = VerticalOffset * -1;
            return finalSize;
        }
    }

}
